/* Dmorte Solitaire Adventure Tool ©2018 Howard A. Kistler */

/* Constants */

const VERSION_NUMBER = '0.9';

const DEBUG_ON = false;

const CELLWIDTH   = 112;
const CELLHEIGHT  = 72;
const CELLPAD     = 14;
const CELLOFFSET  = Math.round(CELLWIDTH / 4);
const BRACKETSIZE = 8;
const UNITWIDTH   = CELLWIDTH + (CELLPAD * 2);
const UNITHEIGHT  = CELLHEIGHT + (CELLPAD * 2);

const CLICKER_SIZE     = 20;
const CLICKER_DELETE_X = CELLWIDTH - (CLICKER_SIZE + 2);
const CLICKER_DELETE_Y = 2;
const CLICKER_ONEWAY_X = 2;
const CLICKER_ONEWAY_Y = CELLHEIGHT - (CLICKER_SIZE + 2);
const CLICKER_TWOWAY_X = CELLWIDTH - (CLICKER_SIZE + 2);
const CLICKER_TWOWAY_Y = CELLHEIGHT - (CLICKER_SIZE + 2);
const CLICKER_PREP_X = 2;
const CLICKER_PREP_Y = 2;

const NODE_BLANK    = '#c0c0c0';
const NODE_START    = '#efefaf';
const NODE_BEGIN    = '#cf5f5f';
const NODE_STANDARD = '#8fefdf';
const NODE_TERMINUS = '#cfafff';
const NODE_PHANTOM  = '#dfdfdf';
const NODE_SELECTED = '#ffdfaf';

const NODEPOINT_START = '#9fff9f';
const NODEPOINT_END   = '#ff9f9f';
const NODEPOINT_BIDI  = '#dfdf7f';

const CARDINAL_NORTH = 'N';
const CARDINAL_SOUTH = 'S';
const CARDINAL_EAST  = 'E';
const CARDINAL_WEST  = 'W';

const CONNECTION_NONE   = 0;
const CONNECTION_ONEWAY = 1;
const CONNECTION_TWOWAY = 2;

const PREP_WIP = 0;
const PREP_COMPLETE = 1;

const JSONPREAMBLE = "data:text/json;charset=utf-8,";

const NEWLINE = '\r\n';

const DEFAULT_LINK_TEXT = 'Go to';

/* Page Reload Interceptor */

window.addEventListener("beforeunload", function(evt){
	var msg = "Are you sure you wish to reload the page?<br/>You will lose any unsaved work.";
	(evt || window.event).returnValue = msg;
	return msg;
});

/* Objects */

function Point(x, y) {
	this.x = x;
	this.y = y;
}

function DesignNode(id, x, y, label, name, content, mappings, prep, style) {
	this.id = id;
	this.x = x
	this.y = y;
	this.label = label;
	this.name = name;
	this.content = content;
	this.mappings = (!Array.isArray(mappings) ? [] : mappings);
	this.prep = (typeof prep != 'undefined' ? prep : PREP_COMPLETE);
	this.style = (typeof style != 'undefined' ? style : NODE_BLANK);
}

/* Variables */

var gridCols = 10;
var gridRows = 12;
var designGrid = [];
var nodeTree = {};

var activeNode = null;

var dragActive = false;
var dragStartX = -1;
var dragStartY = -1;
var isDragging = false;
var dragObject = null;
var isLinking  = false;
var linkType   = CONNECTION_NONE;
var isPicking  = false;

var activeForm = '';

var cframe;
var canvas;
var cnv;

var theNodeEditor;

var projectName;
var projectSeed;
var projectLinkText;

var RNG;

/* Functions */

function initialise() {
	cframe = document.getElementById('canvasframe');

	canvas = document.getElementById("canvas");
	canvas.addEventListener('mousemove',  function(evt) { processMouseMove(evt); });
	canvas.addEventListener('mousedown',  function(evt) { processMouseClick(evt, true); });
	canvas.addEventListener('mouseup',    function(evt) { processMouseClick(evt, false); });
	canvas.addEventListener('mouseleave', function(evt) { processCanvasExit(evt); });

	cnv = canvas.getContext('2d');

	theNodeEditor = document.forms['nodeEditor'];

	for(var gx = 0; gx < gridCols; gx++) {
		designGrid[gx] = [];
		for(var gy = 0; gy < gridRows; gy++) {
			designGrid[gx][gy] = null;
		}
	}

	parseObjectTree();

	RNG = new xoshiro128(projectSeed);

	renderBoard();

	document.getElementById('projtitle').value = projectName;
	setProjectTitle(document.getElementById('projfilename'));
	document.getElementById('projlinktext').value = projectLinkText;
}

function debugMsg(msg, forced) {
	if(DEBUG_ON || true == forced) {
		console.log(msg);
	}
}

function getRandomNodeId() {
	return 'node-' + Math.random().toString(36).substr(2);
}

function processMouseClick(evt, isDown) {
	toggleDragState(evt, isDown);
	if(activeForm != 'Editor') { setForm('Editor'); }
	var mousePos = getMousePosition(evt);
	var clickCell = getGridCell(mousePos.x, mousePos.y);
	if(isPicking) {
		if(designGrid[clickCell.x][clickCell.y] != null) {
			console.log(designGrid[clickCell.x][clickCell.y]);
			var pickedNodeId = designGrid[clickCell.x][clickCell.y];
			if(pickedNodeId != startNode) {
				var oldStart = startNode;
				startNode = pickedNodeId;
				if(oldStart != null) { refreshNode(oldStart); }
				refreshNode(startNode);
				renderBoard();
			}
		}
		isPicking = false;
	}
	else if(isLinking) {
		var makeAForeLink = true;
		var makeABackLink = (linkType == CONNECTION_TWOWAY);
		if(designGrid[clickCell.x][clickCell.y] != null) {
			if(designGrid[clickCell.x][clickCell.y] == activeNode) {
				// cannot link back to self
				makeAForeLink = false;
				makeABackLink = false;
			}
			else {
				var targetNodeId = designGrid[clickCell.x][clickCell.y];
				for(let mp in nodeTree[activeNode].mappings) {
					if(nodeTree[activeNode].mappings[mp].id == targetNodeId) {
						// already mapped, so no forward link
						makeAForeLink = false;
					}
				}
				if(makeABackLink) {
					for(let mpt in nodeTree[targetNodeId].mappings) {
						if(nodeTree[targetNodeId].mappings[mpt].id == activeNode) {
							// already mapped
							makeABackLink = false;
						}
					}
					if(makeABackLink) {
						// create the return mapping for a bidi connection
						addConnection(targetNodeId, activeNode);
					}
				}
				if(makeAForeLink) {
					addConnection(activeNode, targetNodeId);
				}
				renderBoard();
			}
		}
		isLinking = false;
	}
	else if(!isDown) {
		if(isDragging) {
			if(dragObject != null) {
				var startCell = getGridCell(dragStartX, dragStartY);
				if(designGrid[clickCell.x][clickCell.y] == null) {
					designGrid[clickCell.x][clickCell.y] = dragObject;
					nodeTree[dragObject].x = clickCell.x;
					nodeTree[dragObject].y = clickCell.y;
					designGrid[startCell.x][startCell.y] = null;
				}
			}
			isDragging = false;
			dragObject = null;
			dragStartX = -1;
			dragStartY = -1;
			renderBoard();
		}
		else {
			if(designGrid[clickCell.x][clickCell.y] == null) {
				generateDesignNode(getRandomNodeId(), clickCell.x, clickCell.y, 'untitled', '', '', [], PREP_WIP, NODE_PHANTOM);
				renderBoard();
				shownode(designGrid[clickCell.x][clickCell.y]);
			}
			else if(designGrid[clickCell.x][clickCell.y] == activeNode) {
				var localClick = getMousePositionWithinCell(evt);
				if(localClick.x >= CLICKER_DELETE_X && localClick.x <= (CLICKER_DELETE_X + CLICKER_SIZE) && localClick.y >= CLICKER_DELETE_Y && localClick.y <= (CLICKER_DELETE_Y + CLICKER_SIZE)) {
					if(activeNode == startNode) {
						alert('You cannot delete the Start node until you assign another node to be the Start.');
					}
					else {
						if(confirm('Delete this node?')) {
							deleteDesignNode(activeNode);
						}
					}
				}
				else if(localClick.x >= CLICKER_ONEWAY_X && localClick.x <= (CLICKER_ONEWAY_X + CLICKER_SIZE) && localClick.y >= CLICKER_ONEWAY_Y && localClick.y <= (CLICKER_ONEWAY_Y + CLICKER_SIZE)) {
					linkType = CONNECTION_ONEWAY;
					isLinking = true;
					shownode(designGrid[clickCell.x][clickCell.y]);
				}
				else if(localClick.x >= CLICKER_TWOWAY_X && localClick.x <= (CLICKER_TWOWAY_X + CLICKER_SIZE) && localClick.y >= CLICKER_TWOWAY_Y && localClick.y <= (CLICKER_TWOWAY_Y + CLICKER_SIZE)) {
					linkType = CONNECTION_TWOWAY;
					isLinking = true;
					shownode(designGrid[clickCell.x][clickCell.y]);
				}
				else if(localClick.x >= CLICKER_PREP_X && localClick.x <= (CLICKER_PREP_X + CLICKER_SIZE) && localClick.y >= CLICKER_PREP_Y && localClick.y <= (CLICKER_PREP_Y + CLICKER_SIZE)) {
					if(nodeTree[activeNode].prep == PREP_WIP) {
						nodeTree[activeNode].prep = PREP_COMPLETE;
					}
					else if(nodeTree[activeNode].prep == PREP_COMPLETE) {
						nodeTree[activeNode].prep = PREP_WIP;
					}
					shownode(designGrid[clickCell.x][clickCell.y]);
				}
			}
			else {
				shownode(designGrid[clickCell.x][clickCell.y]);
			}
		}
	}
	else {
	}
}

function processMouseMove(evt) {
	checkDragAction(evt);
	if(isDragging) {
		if(dragObject != null) {
			var mousePos = getMousePosition(evt);
			renderBoard();
			drawDesignNode(cnv, nodeTree[dragObject], mousePos.x - (CELLWIDTH / 2), mousePos.y - (CELLHEIGHT / 2));
		}
	}
	if(isLinking) {
	}
}

function processCanvasExit(evt) {
	cancelDragAction(evt);
}

function renderBoard() {
	cnv.clearRect(0, 0, canvas.width, canvas.height);
	drawSnapGrid(cnv);
	drawDesignNodes(cnv);
	document.getElementById('startnodelabel').innerHTML = nodeTree[startNode].label;
}

function getMousePosition(evt) {
	var boundbox = cframe.getBoundingClientRect();
	var x = (evt.clientX - boundbox.left) + cframe.scrollLeft;
	var y = (evt.clientY - boundbox.top) + cframe.scrollTop;
	return new Point(x, y);
}

function getGridCell(x, y) {
	return new Point(Math.floor(x / UNITWIDTH), Math.floor(y / UNITHEIGHT));
}

function getGridCellPoint(x, y) {
	var gx = (x * UNITWIDTH) + CELLPAD;
	var gy = (y * UNITHEIGHT) + CELLPAD;
	return new Point(gx, gy);
}

function getMousePositionWithinCell(evt) {
	var boundbox = cframe.getBoundingClientRect();
	var x = (evt.clientX - boundbox.left) + cframe.scrollLeft;
	var y = (evt.clientY - boundbox.top) + cframe.scrollTop;
	var cell = getGridCell(x, y);
	var localCornerX = (cell.x * UNITWIDTH) + CELLPAD;
	var localCornerY = (cell.y * UNITHEIGHT) + CELLPAD;
	return new Point(x - localCornerX, y - localCornerY);
}

function generateDesignNode(id, x, y, label, name, content, mappings, prep, style) {
	nodeTree[id] = new DesignNode(id, x, y, label, name, content, mappings, prep, style);
	designGrid[x][y] = id;
}

function deleteDesignNode(nodeid) {
	if(typeof nodeTree[nodeid] != undefined && nodeTree[nodeid] != null && nodeTree[nodeid] != '') {
		// remove outbound connections
		for(let m of nodeTree[nodeid].mappings) {
			removeConnection(nodeid, m.id);
		}
		// remove inbound connections
		for(var nkey in nodeTree) {
			var n = nodeTree[nkey];
			for(let nmap of n.mappings) {
				if(nmap.id == nodeid) {
					removeConnection(n.id, nmap.id);
				}
			}
		}
	}
	designGrid[nodeTree[nodeid].x][nodeTree[nodeid].y] = null;
	delete nodeTree[nodeid];
	renderBoard();
}

function getNodeRenderStyle(node) {
	var hasParents = false;
	for(let nkey in nodeTree) {
		if(nkey == node.id) {}
		else {
			if(nodeTree[nkey].mappings.length > 0) {
				for(let m in nodeTree[nkey].mappings) {
					if(nodeTree[nkey].mappings[m].id == node.id) {
						hasParents = true;
					}
				}
			}
		}
	}
	if(node.id == startNode) {
		return NODE_START;
	}
	else if(node.mappings.length > 0) {
		if(hasParents) {
			return NODE_STANDARD;
		}
		else {
			return NODE_BEGIN;
		}
	}
	else {
		if(hasParents) {
			return NODE_TERMINUS;
		}
		else {
			return NODE_PHANTOM;
		}
	}
}

function toggleDragState(evt, dragOn) {
	dragActive = dragOn;
	if(dragActive) {
		var mp = getMousePosition(evt);
		dragStartX = mp.x;
		dragStartY = mp.y;
	}
	else {
		if(!isDragging) {
			dragStartX = -1;
			dragStartY = -1;
		}
	}
}

function checkDragAction(evt) {
	if(dragActive) {
		if(isDragging) {
			var dragCell = getGridCell(dragStartX, dragStartY);
			if(designGrid[dragCell.x][dragCell.y] != null) {
				dragObject = designGrid[dragCell.x][dragCell.y];
			}
			else {
//					isDragging = false;
//					dragActive = false;
			}
		}
		else {
			var mousePos = getMousePosition(evt);
			if(Math.abs(mousePos.x - dragStartX) > 5 || Math.abs(mousePos.y - dragStartY) > 5) {
				isDragging = true;
				var clickCell = getGridCell(mousePos.x, mousePos.y);
				shownode(designGrid[clickCell.x][clickCell.y]);
			}
		}
	}
}

function cancelDragAction(evt) {
	isDragging = false;
	dragObject = null;
	dragStartX = -1;
	dragStartY = -1;
	toggleDragState(evt, false);
	renderBoard();
}

function drawSnapGrid(ctx) {
	var x = 0;
	var y = 0;
	for(var gx = 0; gx < gridCols; gx++) {
		x = (gx * UNITWIDTH) + CELLPAD;
		for(var gy = 0; gy < gridRows; gy++) {
			y = (gy * UNITHEIGHT) + CELLPAD;
			ctx.beginPath();
			ctx.moveTo(x, y);
			ctx.lineTo(x + BRACKETSIZE, y);
			ctx.moveTo(x, y);
			ctx.lineTo(x, y + BRACKETSIZE);
			ctx.moveTo(x + CELLWIDTH, y);
			ctx.lineTo(x + CELLWIDTH - BRACKETSIZE, y);
			ctx.moveTo(x + CELLWIDTH, y);
			ctx.lineTo(x + CELLWIDTH, y + BRACKETSIZE);
			ctx.moveTo(x, y + CELLHEIGHT);
			ctx.lineTo(x + BRACKETSIZE, y + CELLHEIGHT);
			ctx.moveTo(x, y + CELLHEIGHT);
			ctx.lineTo(x, y + CELLHEIGHT - BRACKETSIZE);
			ctx.moveTo(x + CELLWIDTH, y + CELLHEIGHT);
			ctx.lineTo(x + CELLWIDTH - BRACKETSIZE, y + CELLHEIGHT);
			ctx.moveTo(x + CELLWIDTH, y + CELLHEIGHT);
			ctx.lineTo(x + CELLWIDTH, y + CELLHEIGHT - BRACKETSIZE);
			ctx.lineWidth = 1;
			ctx.strokeStyle = "#d0d0d0";
			ctx.stroke();
/*
			ctx.beginPath();
			ctx.moveTo(x + (CELLWIDTH / 2) - BRACKETSIZE, y + (CELLHEIGHT / 2));
			ctx.lineTo(x + (CELLWIDTH / 2) + BRACKETSIZE, y + (CELLHEIGHT / 2));
			ctx.moveTo(x + (CELLWIDTH / 2), y + (CELLHEIGHT / 2) - BRACKETSIZE);
			ctx.lineTo(x + (CELLWIDTH / 2), y + (CELLHEIGHT / 2) + BRACKETSIZE);
			ctx.strokeStyle = "#d0d0d0";
			ctx.stroke();
*/
		}
	}
}

function drawDesignNode(ctx, node, x, y) {
//	debugMsg(node, true);
	var px = x;
	var py = y;
	if(isNaN(x) || isNaN(y)) {
		var gc = getGridCellPoint(node.x, node.y);
		debugMsg(gc);
		px = gc.x;
		py = gc.y;
	}
	cx = px + (CELLWIDTH / 2);
	cy = py + (CELLHEIGHT / 2);
	ctx.beginPath();
	ctx.strokeStyle = "#000000";
	if(node.prep == PREP_WIP) {
		ctx.setLineDash([4, 4]);
	}
	ctx.fillStyle = (node.id == activeNode ? NODE_SELECTED : node.style);
	ctx.rect(px, py, CELLWIDTH, CELLHEIGHT);
	ctx.fill();
	ctx.stroke();
	ctx.setLineDash([]);
	ctx.beginPath();
	ctx.strokeStyle = "#ffffff";
	ctx.fillStyle = "#000000";
	ctx.font = '18px serif';
	ctx.textAlign = 'center';
	ctx.textBaseline = 'middle';
	ctx.fillText(node.label, cx, cy, CELLWIDTH - 8);
	if(node.id == activeNode && !isDragging) {
		// draw deletion clickable gadget
		var gadx = px + CLICKER_DELETE_X;
		var gady = py + CLICKER_DELETE_Y;
		var orix = gadx + (CLICKER_SIZE / 2);
		var oriy = gady + (CLICKER_SIZE / 2);
		ctx.beginPath();
		ctx.lineWidth = 3;
		ctx.strokeStyle = "#c06060";
		ctx.moveTo(orix - 5, oriy - 5);
		ctx.lineTo(orix + 5, oriy + 5);
		ctx.moveTo(orix + 5, oriy - 5);
		ctx.lineTo(orix - 5, oriy + 5);
		ctx.stroke();
		ctx.lineWidth = 1;
		// draw one-way connection clickable gadget
		gadx = px + CLICKER_ONEWAY_X;
		gady = py + CLICKER_ONEWAY_Y;
		orix = gadx + (CLICKER_SIZE / 2);
		oriy = gady + (CLICKER_SIZE / 2);
/*
		ctx.beginPath();
		ctx.strokeStyle = "#808080";
		ctx.fillStyle = "#ffffff";
		ctx.rect(gadx, gady, CLICKER_SIZE, CLICKER_SIZE);
		ctx.fill();
		ctx.stroke();
*/
		ctx.beginPath();
		var poly = [
			new Point(-8, -6),
			new Point( 8, -6),
			new Point( 0,  8)
		];
		renderPolygon(ctx, poly, orix, oriy, ((isLinking && (linkType == CONNECTION_ONEWAY)) ? "#c06060" : "#808080"), "#808080");
		ctx.beginPath();
		ctx.lineWidth = 2;
		ctx.strokeStyle = "#ffffff";
		ctx.moveTo(orix, oriy - 5);
		ctx.lineTo(orix, oriy + 5);
		ctx.moveTo(orix - 4, oriy);
		ctx.lineTo(orix + 4, oriy);
		ctx.stroke();
		// draw two-way connection clickable gadget
		gadx = px + CLICKER_TWOWAY_X;
		gady = py + CLICKER_TWOWAY_Y;
		orix = gadx + (CLICKER_SIZE / 2);
		oriy = gady + (CLICKER_SIZE / 2);
/*
		ctx.beginPath();
		ctx.lineWidth = 1;
		ctx.strokeStyle = "#808080";
		ctx.fillStyle = "#ffffff";
		ctx.rect(gadx, gady, 20, 20);
		ctx.fill();
		ctx.stroke();
*/		ctx.beginPath();
		ctx.strokeStyle = "#808080";
		ctx.fillStyle = ((isLinking && (linkType == CONNECTION_TWOWAY)) ? "#c06060" : "#808080");
		ctx.arc(orix, oriy, 6, 0, Math.PI*2);
		ctx.fill();
		ctx.stroke();
		ctx.beginPath();
		ctx.lineWidth = 2;
		ctx.strokeStyle = "#ffffff";
		ctx.moveTo(orix, oriy - 5);
		ctx.lineTo(orix, oriy + 5);
		ctx.moveTo(orix - 5, oriy);
		ctx.lineTo(orix + 5, oriy);
		ctx.stroke();
		// draw prep clickable gadget
		var gadx = px + CLICKER_PREP_X;
		var gady = py + CLICKER_PREP_Y;
		var orix = gadx + (CLICKER_SIZE / 2);
		var oriy = gady + (CLICKER_SIZE / 2);
		ctx.beginPath();
		ctx.strokeStyle = "#808080";
		ctx.arc(orix, oriy, 6, 0, Math.PI*2);
		ctx.lineWidth = 2;
		ctx.stroke();
		if(node.prep == PREP_COMPLETE) {
			ctx.beginPath();
			ctx.arc(orix, oriy, 4, 0, Math.PI*2);
			ctx.lineWidth = 1;
			ctx.fillStyle = "#808080";
			ctx.fill();
		}
		ctx.lineWidth = 1;
		// show active linking message
		if(isLinking) {
			ctx.beginPath();
			ctx.rect(px + 16, py + 26, CELLWIDTH - 32, CELLHEIGHT - 52);
			ctx.fillStyle = "#ffffff";
			ctx.fill();
			ctx.strokeStyle = "#808080";
			ctx.stroke();
			ctx.beginPath();
			ctx.strokeStyle = "#903030";
			ctx.fillStyle = "#903030";
			ctx.font = '14px serif';
			ctx.textAlign = 'center';
			ctx.textBaseline = 'middle';
			ctx.fillText("Link To", cx, cy, CELLWIDTH - 8);
		}
	}
}

function drawDesignNodes(ctx) {
	// draw nodes
	for(let nkey in nodeTree) {
		drawDesignNode(ctx, nodeTree[nkey]);
	}
	// draw connectors
	for(let nkey in nodeTree) {
		nodeObj = nodeTree[nkey];
		nodeStart = nodeTree[nkey];
		for(let m of nodeObj.mappings) {
			nodeEnd = nodeTree[m.id];
			nodeTarget = nodeTree[m.id];
			bidi = false;
			for(let mt of nodeTarget.mappings) {
				if(mt.id == nkey) {
					bidi = true;
				}
			}
			if(bidi && (nodeEnd.y > nodeStart.y)) { /* should already be connected from the upper to lower in previous connection */ }
			else {
				connectNodes(ctx, nodeStart.x, nodeStart.y, nodeEnd.x, nodeEnd.y, bidi);
			}
		}
	}
}

function shownode(id) {
	activeNode = id;
	renderBoard();
	// clear mappings first
	var connList = document.getElementById('connectionList');
	connList.innerHTML = '';
	// get the node and display the contents
	node = nodeTree[id];
	theNodeEditor.currnode.value = id;
	theNodeEditor.nodelabel.value = node.label;
	theNodeEditor.nodename.value = node.name;
	theNodeEditor.nodecontent.value = node.content;
	var connectionTable = document.createElement('table');
	connectionTable.class = 'mappingTable';
	var connectionTableContents = '';
	for(let mp of node.mappings) {
		connectionTableContents +=
			'<tr>' +
				'<td class="mappingItemDelete"><a href="" onclick="return deleteConnection(' + node.mappings.indexOf(mp) + ');">[X]</a></td>' +
				'<td class="mappingItemConnection">' + nodeTree[mp.id].label + '</td>' +
				'<td class="mappingItemContent"><input type="text" class="formgadgetinline" size="36" value="' + mp.desc + '" oninput="updateConnection(' + node.mappings.indexOf(mp) + ',this);" onchange="updateConnection(' + node.mappings.indexOf(mp) + ',this);"/></td>' +
			'</tr>';
	}
	connectionTable.innerHTML = connectionTableContents;
	connList.appendChild(connectionTable);
}

function setForm(formname) {
	document.getElementById('tabProject').className = "tab tabInactive";
	document.getElementById('tabEditor').className = "tab tabInactive";
	document.getElementById('tabReview').className = "tab tabInactive";
	document.getElementById('tabProduce').className = "tab tabInactive";
	document.getElementById('tabHelp').className = "tab tabInactive";
	document.getElementById('tab' + formname).className = "tab tabActive";
	document.getElementById('formProject').style.display = "none";
	document.getElementById('formEditor').style.display = "none";
	document.getElementById('formReview').style.display = "none";
	document.getElementById('formProduce').style.display = "none";
	document.getElementById('formHelp').style.display = "none";
	document.getElementById('form' + formname).style.display = "block";
	activeForm = formname;
	switch(formname) {
		case 'Review':
			renderReviewText();
			break;
		case 'Produce':
			renderProducedText();
			break;
		default:
			break;
	}
}

function renderPolygon(ctx, poly, ctrx, ctry, fillStyle, outlineStyle) {
	ctx.beginPath();
	ctx.moveTo(poly[0].x + ctrx, poly[0].y + ctry);
	for(var p = 1; p < poly.length; p++) {
		ctx.lineTo(poly[p].x + ctrx, poly[p].y + ctry);
	}
	ctx.closePath();
	if(typeof fillStyle != undefined && fillStyle != null) {
		ctx.fillStyle = fillStyle;
		ctx.fill();
	}
	if(typeof outlineStyle != undefined && outlineStyle != null) {
		ctx.strokeStyle = outlineStyle;
		ctx.stroke();
	}
}

function drawPolygon(ctx, poly, ctrx, ctry, style) {
	renderPolygon(ctx, poly, ctrx, ctry, style, false);
}

function fillPolygon(ctx, poly, ctrx, ctry, style) {
	renderPolygon(ctx, poly, ctrx, ctry, style, true);
}

function plotNodelineUni(ctx, x, y, vshift, hshift, orientation, style) {
	var gc = getGridCellPoint(x, y);
	var orix = gc.x + hshift;
	var oriy = gc.y + vshift;
	ctx.beginPath();
	var poly = [];
	switch(orientation) {
		case CARDINAL_NORTH:
			poly = [
				new Point(-8, -6),
				new Point( 8, -6),
				new Point( 0,  8)
			];
			break;
		case CARDINAL_SOUTH:
			poly = [
				new Point(-8,  6),
				new Point( 8,  6),
				new Point( 0, -8)
			];
			break;
		case CARDINAL_EAST:
			poly = [
				new Point(-7, -8),
				new Point(-7,  8),
				new Point( 7,  0)
			];
			break;
		case CARDINAL_WEST:
			poly = [
				new Point( 7, -8),
				new Point( 7,  8),
				new Point(-7,  0)
			];
			break;
		default:
			poly = [
				new Point(-8,  6),
				new Point( 8,  6),
				new Point( 0, -8)
			];
			break;
	}
	renderPolygon(ctx, poly, orix, oriy, style, "#000000");
}

function plotNodelineBidi(ctx, x, y, vshift, hshift) {
	var gc = getGridCellPoint(x, y);
	var orix = gc.x + hshift;
	var oriy = gc.y + vshift;
	ctx.beginPath();
	ctx.strokeStyle = "#000000";
	ctx.fillStyle   = NODEPOINT_BIDI;
	ctx.arc(orix, oriy, 6, 0, Math.PI*2);
	ctx.fill();
	ctx.stroke();
}

function connectNodes(ctx, fx, fy, tx, ty, bidi) {
	if(fy == ty) {
		// horizontal connector
		offset = -(CELLHEIGHT / 4);
		vshift = (CELLHEIGHT / 2);
		hshift = CELLWIDTH;
		bshift = -7;
		orient = CARDINAL_EAST;
		if(bidi == true) {
			offset = 0;
			bshift = -6;
		}
		if(tx < fx) {
			offset = -offset;
			hshift = 0;
			bshift = -bshift;
			orient = CARDINAL_WEST;
		}
		var fgc = getGridCellPoint(fx, fy);
		fcx = fgc.x + hshift;
		fcy = fgc.y + vshift + offset;
		mcx = fcx + (((tcx - fcx) / 2));
		var tgc = getGridCellPoint(tx, ty);
		tcx = tgc.x + CELLWIDTH - hshift;
		tcy = tgc.y + vshift + offset;
		if(bidi != undefined && bidi == true) {
			plotNodelineBidi(ctx, fx, fy, vshift, hshift);
			plotNodelineBidi(ctx, tx, ty, vshift, CELLWIDTH - hshift);
			ctx.strokeStyle = "#b8b86f";
		}
		else {
			plotNodelineUni(ctx, fx, fy, vshift + offset, hshift, orient, NODEPOINT_START);
			plotNodelineUni(ctx, tx, ty, vshift + offset, CELLWIDTH - hshift, orient, NODEPOINT_END);
			ctx.strokeStyle = "#404040";
		}
		ctx.beginPath();
		ctx.moveTo(fcx - bshift, fcy);
		ctx.lineTo(tcx + bshift, tcy);
		ctx.stroke();
	}
	else {
		// vertical connector
		offset = -CELLOFFSET;
		vshift = 0;
		hshift = CELLWIDTH / 2;
		tshift = 8;
		bshift = -6;
		orient = CARDINAL_NORTH;
		if(bidi == true) {
			offset = 0;
			bshift = -8;
		}
		if(ty < fy) {
			offset = -offset;
			vshift = CELLHEIGHT;
			tshift = -tshift;
			bshift = -bshift;
			orient = CARDINAL_SOUTH;
		}
		var fgc = getGridCellPoint(fx, fy);
		fcx = fgc.x + (CELLWIDTH / 2) + offset;
		fcy = fgc.y + CELLHEIGHT;
		var tgc = getGridCellPoint(tx, ty);
		tcx = tgc.x + (CELLWIDTH / 2) + offset;
		tcy = tgc.y;
		mcy = fcy + (((tcy - fcy) / 2));
		if(bidi != undefined && bidi == true) {
			plotNodelineBidi(ctx, fx, fy, CELLHEIGHT - vshift, hshift);
			plotNodelineBidi(ctx, tx, ty, vshift, hshift);
			ctx.strokeStyle = "#b8b86f";
		}
		else {
			plotNodelineUni(ctx, fx, fy, CELLHEIGHT - vshift, hshift + offset, orient, NODEPOINT_START);
			plotNodelineUni(ctx, tx, ty, vshift, hshift + offset, orient, NODEPOINT_END);
			ctx.strokeStyle = "#404040";
		}
		ctx.beginPath();
		ctx.moveTo(fcx, (fcy + tshift - vshift));
		ctx.lineTo(fcx, mcy);
		ctx.lineTo(tcx, mcy);
		ctx.lineTo(tcx, (tcy + bshift + vshift));
		ctx.stroke();
	}
}

function clearProject() {
	for(var x = 0; x < designGrid.length; x++) {
		for(var y = 0; y < designGrid[x].length; y++) {
			designGrid[x][y] = null;
		}
	}
	nodeTree.length = 0;
	nodeTree = {};
}

function newProject() {
	var newProjOk = true;
	if(Object.keys(nodeTree).length > 0) {
		newProjOk = confirm('Creating a new project will clear the current one, and you will lose any unsaved data. Is this okay?');
	}
	if(newProjOk) {
		clearProject();
		projectTitle = '[blank]';
		seedValue = new Date().getTime();
		saveFilename = 'default';
		document.getElementById('projtitle').value = projectTitle;
		document.getElementById('projfilename').value = saveFilename;
		document.getElementById('projlinktext').value = DEFAULT_LINK_TEXT;
		startNode = getRandomNodeId();
		generateDesignNode(startNode, 0, 0, 'START', '', '', [], PREP_COMPLETE, NODE_START);
		renderBoard();
		shownode(designGrid[0][0]);
	}
}

function loadProject(cmp) {
	var loadFile = cmp.files[0];
	var lastPeriod = loadFile.name.lastIndexOf(".");
	var extn = (lastPeriod < 0 ? "null" : loadFile.name.substr(lastPeriod).toLowerCase());
	if(".dmrt" != extn) {
//				document.getElementById("loadmsg").innerHTML = "<span class='warning'><b>" + loadFile.name + "</b> does not have <b>.dmrt</b> extension</span>";
	}
	else {
//				document.getElementById("loadmsg").innerHTML = "<span class='success'>Loading map <b>" + loadFile.name + "</b> ...</span>";
		var filereader = new FileReader();
		filereader.onload = function(lf) {
			clearProject();
			parseObjectTree(lf.target.result.trim());
			renderBoard();
			document.getElementById('projtitle').value = projectName;
			document.getElementById('projlinktext').value = projectLinkText;
		}
		filereader.readAsText(loadFile);
		document.getElementById('projfilename').value = loadFile.name.substr(0, lastPeriod);
	}
}

function saveProject() {
	var mapname = 'example';
	var projectTitle = document.getElementById('projtitle').value;
	var seedValue = projectSeed;
	var saveFilename = document.getElementById('projfilename').value;
	var saveLinkText = document.getElementById('projlinktext').value;
	if(!projectTitle) { projectTitle = '[blank]'; }
	if(!seedValue) { seedValue = new Date().getTime(); }
	if(!saveFilename) { saveFilename = 'default'; }

	var saveFile = {};
	saveFile['name'] = projectTitle;
	saveFile['seed'] = seedValue;
	saveFile['start'] = startNode;
	saveFile['linktext'] = saveLinkText;
	saveFile['nodetree'] = nodeTree;
	var nodeTreeToJson = JSON.stringify(saveFile);
	var designGridToJson = JSON.stringify(designGrid);

	// get URI for file
	var jsonUrl = JSONPREAMBLE + encodeURIComponent(nodeTreeToJson);

	// pseudolink for named download
	var pseudoanchor = document.createElement("a");
	pseudoanchor.href = jsonUrl;
	pseudoanchor.download = saveFilename + ".dmrt";
	pseudoanchor.style.display = "none";
	document.body.appendChild(pseudoanchor);
	pseudoanchor.click();
	document.body.removeChild(pseudoanchor);
}

function setStartNode() {
	isPicking = true;
}

function parseObjectTree(src) {
	if(typeof src != undefined && src != null) {
		arrayOfData = JSON.parse(src);
	}
	else {
		arrayOfData = JSON.parse(document.getElementById('jsonnodetree').innerHTML);
	}
	projectName = arrayOfData.name;
	projectSeed = (typeof arrayOfData.seed != 'undefined' ? arrayOfData.seed : new Date().getTime());
	projectLinkText = (typeof arrayOfData.linktext != 'undefined' ? arrayOfData.linktext : DEFAULT_LINK_TEXT);
	nodeSet = arrayOfData.nodetree;
	startNode = arrayOfData.start;

	for(var key in nodeSet) {
		var n = nodeSet[key];
		generateDesignNode(n.id, n.x, n.y, n.label, n.name, n.content, n.mappings, n.prep);
	}

	for(let nkey in nodeTree) {
		nodeTree[nkey].style = getNodeRenderStyle(nodeTree[nkey]);
	}
}

function refreshNode(nodeId) {
	nodeTree[nodeId].style = getNodeRenderStyle(nodeTree[nodeId]);
}

function updateNodeLabel(cmp) {
	if(theNodeEditor.currnode.value != '') {
		nodeTree[theNodeEditor.currnode.value].label = cmp.value;
		renderBoard();
	}
}

function updateNodeName(cmp) {
	if(theNodeEditor.currnode.value != '') {
		nodeTree[theNodeEditor.currnode.value].name = cmp.value;
	}
}

function updateNodeContent(cmp) {
	if(theNodeEditor.currnode.value != '') {
		nodeTree[theNodeEditor.currnode.value].content = cmp.value;
	}
}

function updateConnection(mapid, cmp) {
	if(theNodeEditor.currnode.value != '') {
		if(Array.isArray(nodeTree[theNodeEditor.currnode.value].mappings)) {
			if(typeof nodeTree[theNodeEditor.currnode.value].mappings[mapid] != undefined) {
				nodeTree[theNodeEditor.currnode.value].mappings[mapid].desc = cmp.value;
			}
		}
	}
}

function removeConnection(nodeid, mappingid) {
	if(typeof nodeTree[nodeid] != undefined) {
		var node = nodeTree[nodeid];
		if(Array.isArray(node.mappings)) {
			for(let mp of node.mappings) {
				if(mp.id == mappingid) {
					var nindex = node.mappings.indexOf(mp);
					if(nindex > -1) {
						node.mappings.splice(nindex, 1);
					}
				}
			}
		}
	}
}

function deleteConnection(mapid) {
	if(confirm('Delete this connection?')) {
		if(theNodeEditor.currnode.value != '') {
			if(typeof nodeTree[theNodeEditor.currnode.value] != undefined) {
				var node = nodeTree[theNodeEditor.currnode.value];
				if(Array.isArray(node.mappings)) {
					if(typeof node.mappings[mapid] != undefined) {
						var mappingId = node.mappings[mapid].id;
						// delete the mapping itself
						node.mappings.splice(mapid, 1);
						refreshNode(mappingId);
						refreshNode(theNodeEditor.currnode.value);
						renderBoard();
						shownode(theNodeEditor.currnode.value);
					}
				}
			}
		}
	}
	return false;
}

function addConnection(sourceNodeId, targetNodeId) {
	if(Array.isArray(nodeTree[sourceNodeId].mappings)) {
		var newConn = {
			'id': targetNodeId,
			'desc': '[none]'
		};
		nodeTree[sourceNodeId].mappings.push(newConn);
		refreshNode(sourceNodeId);
		refreshNode(targetNodeId);
	}
}

function setProjectTitle(cmp) {
	var projectTitle = document.getElementById('projtitle').value;
	var filenameField = document.getElementById('projfilename');
	var regex = /[^a-zA-Z0-9_]/gi;
	projectTitle = projectTitle.replace(regex, '');
	filenameField.value = projectTitle;
}

function setProjectFilename(cmp) {
	var filenameField = document.getElementById('projfilename');
	var projectFilename = filenameField.value;
	var regex = /[^a-zA-Z0-9_]/gi;
	projectFilename = projectFilename.replace(regex, '');
	filenameField.value = projectFilename;
}

function goToHelp(sect) {
	var helpdiv = document.getElementById('formHelp');
	var helpsect = document.getElementById(sect);
	helpdiv.scrollTop = helpsect.offsetTop - helpdiv.offsetTop;
	return false;
}

function regenerateSeed(refresh) {
	var newseed = new Date().getTime().toString();
	newseed = newseed.split('').reverse().join('');
	var saveseed = parseInt(newseed)
	projectSeed = saveseed;
	if(refresh) {
		renderProducedText();
	}
}

function renderReviewText() {
	document.getElementById('reviewBox').value = getReviewText();
}

function getReviewText() {
	var fullSet = traverseNodeTree();
	console.log(fullSet);
	var traversalSet = fullSet['mapped'];
	var orphanedSet  = fullSet['orphaned'];
	console.log(traversalSet.length);
	console.log(orphanedSet.length);
	var displayText = getReviewNodeBlock(startNode) + NEWLINE;
	for(nodeIndex in traversalSet) {
		if(traversalSet[nodeIndex] != startNode) {
			displayText = displayText + getReviewNodeBlock(traversalSet[nodeIndex]) + NEWLINE;
		}
	}
	if(orphanedSet.length > 0) {
		displayText = displayText + "==== UNREACHABLE NODES ====" + NEWLINE + NEWLINE;
		for(nodeIndex in orphanedSet) {
			displayText = displayText + getReviewNodeBlock(orphanedSet[nodeIndex]) + NEWLINE;
		}
	}
	return displayText;
}

function traverseNodeTree() {
	var traversedTree  = [];
	var traversedNodes = [];
	var orphanedNodes  = [];
	// set START node at beginning of traversal tree
	traversedNodes.push(startNode);
	// add any other BEGIN nodes at start
	for(nodeId in nodeTree) {
		if(getNodeRenderStyle(nodeTree[nodeId]) == NODE_BEGIN) {
			traversedNodes.push(nodeId);
		}
	}
	// traverse remaining tree
	traverseNode(startNode, traversedNodes);
	traversedTree['mapped'] = traversedNodes;
	for(nodeId in nodeTree) {
		if(!traversedNodes.includes(nodeId)) {
			orphanedNodes.push(nodeId);
		}
	}
	traversedTree['orphaned'] = orphanedNodes;
	return traversedTree;
}

function traverseNode(nodeId, nodeArray) {
	var node = nodeTree[nodeId];
	for(mapId in node['mappings']) {
		mappedNode = nodeTree[node['mappings'][mapId]['id']];
		mappedNodeId = mappedNode.id;
		if(!nodeArray.includes(mappedNodeId)) {
			nodeArray.push(mappedNodeId);
			traverseNode(mappedNodeId, nodeArray);
		}
	}
}

function exportReviewText() {
	var revText = getReviewText();
	var mapname = 'default';
	if(document.getElementById('projfilename').value > '') {
		mapname = document.getElementById('projfilename').value;
	}
	var pseudoanchor = document.createElement("a");
	var txtFile = new Blob([revText], {type: 'text/plain'});
	pseudoanchor.href = URL.createObjectURL(txtFile);
	pseudoanchor.download = mapname + "-review.txt";
	pseudoanchor.style.display = "none";
	document.body.appendChild(pseudoanchor);
	pseudoanchor.click();
	document.body.removeChild(pseudoanchor);
}

function getReviewNodeBlock(nodeId) {
	var node = nodeTree[nodeId];
	var nodeText = node['name'] + ' [' + node['label'] + ']' + NEWLINE;
	nodeText = nodeText + node['content'] + NEWLINE;
	for(mapId in node['mappings']) {
		mappedNode = nodeTree[node['mappings'][mapId]['id']];
		nodeText = nodeText + '» ' + mappedNode['name'] + ': ' + node['mappings'][mapId]['desc'] + NEWLINE;
	}
	return nodeText;
}

function renderProducedText() {
	document.getElementById('produceBox').value = getProducedText();
}

function getProducedText() {
	RNG.seedRNG(projectSeed);
	var beginNodeSet = [];
	for(nodeId in nodeTree) {
		if(getNodeRenderStyle(nodeTree[nodeId]) == NODE_BEGIN) {
			beginNodeSet.push(nodeId);
		}
	}	
	var randomNodeSeq = [];
	for(nodeId in nodeTree) {
		if(nodeId != startNode && !beginNodeSet.includes(nodeId)) {
			randomNodeSeq.push(nodeId);
		}
	}
	var arrayPtr = randomNodeSeq.length;
	while(arrayPtr > 0) {
		randomPtr = RNG.rnd(arrayPtr);
		arrayPtr--;
		placeholder = randomNodeSeq[arrayPtr];
		randomNodeSeq[arrayPtr] = randomNodeSeq[randomPtr];
		randomNodeSeq[randomPtr] = placeholder;
	}
	randomNodeSeq.splice(0, 0, startNode);
	for(var i = 0; i < beginNodeSet.length; i++) {
		randomNodeSeq.splice(i + 1, 0, beginNodeSet[i]);
	}
	var headNodeCount = 1 + beginNodeSet.length;
	var displayText = '';
	var linkText = document.getElementById('projlinktext').value;
	for(rndId in randomNodeSeq) {
		var rndKey = randomNodeSeq[rndId];
		var paragraphNum = randomNodeSeq.indexOf(rndKey) + 1;
		var node = nodeTree[rndKey];
		var nodeText = paragraphNum + ' - ' + node['name'] + NEWLINE;
		nodeText = nodeText + node['content'] + NEWLINE;
		for(mapId in node['mappings']) {
			var mapKey = node['mappings'][mapId]['id'];
			var paragraphLinkNum = randomNodeSeq.indexOf(mapKey) + 1;
			nodeText = nodeText + node['mappings'][mapId]['desc'] + ' ' + linkText + ' ' +  paragraphLinkNum + NEWLINE;
		}
		displayText = displayText + nodeText + NEWLINE;
	}
	return displayText;
}

function exportProducedText() {
	var prodText = getProducedText();
	var mapname = 'default';
	if(document.getElementById('projfilename').value > '') {
		mapname = document.getElementById('projfilename').value;
	}
	var pseudoanchor = document.createElement("a");
	var txtFile = new Blob([prodText], {type: 'text/plain'});
	pseudoanchor.href = URL.createObjectURL(txtFile);
	pseudoanchor.download = mapname + "-produced.txt";
	pseudoanchor.style.display = "none";
	document.body.appendChild(pseudoanchor);
	pseudoanchor.click();
	document.body.removeChild(pseudoanchor);
}

function writeVersionNumber() {
	document.write('(v' + VERSION_NUMBER + ')');
}